import asyncio
import unittest
from greenletio import spawn
from greenletio.core import bridge
from greenletio.green import socket


class TestSocket(unittest.TestCase):
    def setUp(self):
        bridge.reset()

    def tearDown(self):
        bridge.stop()

    def test_sendall_recv(self):
        var = None

        def server():
            server_socket = socket.socket()
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(('127.0.0.1', 7000))
            server_socket.listen(5)
            conn, _ = server_socket.accept()
            data = conn.recv(1024)
            conn.sendall(data.upper())
            conn.close()
            server_socket.close()

        def client():
            nonlocal var
            client_socket = socket.socket()
            client_socket.connect(('127.0.0.1', 7000))
            client_socket.sendall(b'hello')
            var = client_socket.recv(1024)
            client_socket.close()

        async def main():
            nonlocal var
            spawn(server)
            spawn(client)
            while var is None:
                await asyncio.sleep(0)

        asyncio.get_event_loop().run_until_complete(main())
        assert var == b'HELLO'

    def test_sendto_recvfrom(self):
        var = None

        def server():
            server_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(('127.0.0.1', 7000))
            data, addr = server_socket.recvfrom(1024)
            server_socket.sendto(data.upper(), addr)
            server_socket.close()

        def client():
            nonlocal var
            client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            client_socket.sendto(b'hello', ('127.0.0.1', 7000))
            var, addr = client_socket.recvfrom(1024)
            client_socket.close()

        async def main():
            nonlocal var
            spawn(server)
            spawn(client)
            while var is None:
                await asyncio.sleep(0)

        asyncio.get_event_loop().run_until_complete(main())
        assert var == b'HELLO'



import errno
import os
from greenletio.io import wait_to_read, wait_to_write
from greenletio.patcher import copy_globals
from socket import _GLOBAL_DEFAULT_TIMEOUT, SOCK_STREAM, AF_INET, AF_INET6, \
    IPPROTO_IPV6, IPV6_V6ONLY, SOL_SOCKET, SO_REUSEADDR, SO_REUSEPORT, \
    error, getaddrinfo, _socket, has_ipv6
try:
    from socket import has_dualstack_ipv6
except ImportError:  # pragma: no cover
    pass
import socket as _original_socket_

copy_globals(_original_socket_, globals())


class socket(_original_socket_.socket):
    def _nonblocking_read(self, method, *args, **kwargs):
        self.setblocking(False)
        while True:
            try:
                ret = method(*args, **kwargs)
                break
            except (OSError, BlockingIOError) as exc:
                err = exc.errno
                if err in (errno.EAGAIN, errno.EWOULDBLOCK):
                    wait_to_read(self.fileno())
                else:  # pragma: no cover
                    raise
        return ret

    def _nonblocking_write(self, method, *args, **kwargs):
        self.setblocking(False)
        while True:
            try:
                ret = method(*args, **kwargs)
                break
            except (OSError, BlockingIOError) as exc:
                err = exc.errno
                ret = None
                if err in (errno.EAGAIN, errno.EWOULDBLOCK, errno.EINPROGRESS):
                    wait_to_write(self.fileno())
                    if err == errno.EINPROGRESS:  # pragma: no cover
                        break
                else:  # pragma: no cover
                    raise
        return ret

    def accept(self, *args, **kwargs):
        wait_to_read(self.fileno())
        conn, address = self._nonblocking_read(super().accept, *args, **kwargs)
        conn.setblocking(False)
        fd = conn.detach()
        return socket(fileno=fd), address

    def connect(self, *args, **kwargs):
        return self._nonblocking_write(super().connect, *args, **kwargs)

    def connect_ex(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_write(super().connect_ex, *args, **kwargs)

    def recv(self, *args, **kwargs):
        return self._nonblocking_read(super().recv, *args, **kwargs)

    def recvfrom(self, *args, **kwargs):
        return self._nonblocking_read(super().recvfrom, *args, **kwargs)

    def recvmsg(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_read(super().recvmsg, *args, **kwargs)

    def recvmsg_into(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_read(super().recvmsg_into, *args, **kwargs)

    def recvfrom_into(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_read(super().recvfrom_into, *args, **kwargs)

    def recv_into(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_read(super().recv_into, *args, **kwargs)

    def send(self, *args, **kwargs):
        return self._nonblocking_write(super().send, *args, **kwargs)

    def sendall(self, data, flags=0):
        tail = self.send(data, flags)
        len_data = len(data)
        while tail < len_data:  # pragma: no cover
            tail += self.send(data[tail:], flags)

    def sendto(self, *args, **kwargs):
        return self._nonblocking_write(super().sendto, *args, **kwargs)

    def sendmsg(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_write(super().sendmsg, *args, **kwargs)

    def sendmsg_afalg(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_write(super().sendmsg_afalg, *args, **kwargs)

    def sendfile(self, *args, **kwargs):  # pragma: no cover
        raise RuntimeError('socket.sendfile is not supported')


# The create_connection and create_server functions below are identical copies
# to those in the Python 3.8. They are included here to ensure they
# instantiate the green versions of the socket class.

def create_connection(address, timeout=_GLOBAL_DEFAULT_TIMEOUT,
                      source_address=None):  # pragma: no cover
    """Connect to *address* and return the socket object.
    Convenience function.  Connect to *address* (a 2-tuple ``(host,
    port)``) and return the socket object.  Passing the optional
    *timeout* parameter will set the timeout on the socket instance
    before attempting to connect.  If no *timeout* is supplied, the
    global default timeout setting returned by :func:`getdefaulttimeout`
    is used.  If *source_address* is set it must be a tuple of (host, port)
    for the socket to bind as a source address before making the connection.
    A host of '' or port 0 tells the OS to use the default.
    """

    host, port = address
    err = None
    for res in getaddrinfo(host, port, 0, SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        sock = None
        try:
            sock = socket(af, socktype, proto)
            if timeout is not _GLOBAL_DEFAULT_TIMEOUT:
                sock.settimeout(timeout)
            if source_address:
                sock.bind(source_address)
            sock.connect(sa)
            # Break explicitly a reference cycle
            err = None
            return sock

        except error as _:
            err = _
            if sock is not None:
                sock.close()

    if err is not None:
        try:
            raise err
        finally:
            # Break explicitly a reference cycle
            err = None
    else:
        raise error("getaddrinfo returns an empty list")


def create_server(address, *, family=AF_INET, backlog=None, reuse_port=False,
                  dualstack_ipv6=False):  # pragma: no cover
    """Convenience function which creates a SOCK_STREAM type socket
    bound to *address* (a 2-tuple (host, port)) and return the socket
    object.
    *family* should be either AF_INET or AF_INET6.
    *backlog* is the queue size passed to socket.listen().
    *reuse_port* dictates whether to use the SO_REUSEPORT socket option.
    *dualstack_ipv6*: if true and the platform supports it, it will
    create an AF_INET6 socket able to accept both IPv4 or IPv6
    connections. When false it will explicitly disable this option on
    platforms that enable it by default (e.g. Linux).
    >>> with create_server(('', 8000)) as server:
    ...     while True:
    ...         conn, addr = server.accept()
    ...         # handle new connection
    """
    if reuse_port and not hasattr(_socket, "SO_REUSEPORT"):
        raise ValueError("SO_REUSEPORT not supported on this platform")
    if dualstack_ipv6:
        if not has_dualstack_ipv6():
            raise ValueError("dualstack_ipv6 not supported on this platform")
        if family != AF_INET6:
            raise ValueError("dualstack_ipv6 requires AF_INET6 family")
    sock = socket(family, SOCK_STREAM)
    try:
        # Note about Windows. We don't set SO_REUSEADDR because:
        # 1) It's unnecessary: bind() will succeed even in case of a
        # previous closed socket on the same address and still in
        # TIME_WAIT state.
        # 2) If set, another socket is free to bind() on the same
        # address, effectively preventing this one from accepting
        # connections. Also, it may set the process in a state where
        # it'll no longer respond to any signals or graceful kills.
        # See: msdn2.microsoft.com/en-us/library/ms740621(VS.85).aspx
        if os.name not in ('nt', 'cygwin') and \
                hasattr(_socket, 'SO_REUSEADDR'):
            try:
                sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
            except error:
                # Fail later on bind(), for platforms which may not
                # support this option.
                pass
        if reuse_port:
            sock.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
        if has_ipv6 and family == AF_INET6:
            if dualstack_ipv6:
                sock.setsockopt(IPPROTO_IPV6, IPV6_V6ONLY, 0)
            elif hasattr(_socket, "IPV6_V6ONLY") and \
                    hasattr(_socket, "IPPROTO_IPV6"):
                sock.setsockopt(IPPROTO_IPV6, IPV6_V6ONLY, 1)
        try:
            sock.bind(address)
        except error as err:
            msg = '%s (while attempting to bind on address %r)' % \
                (err.strerror, address)
            raise error(err.errno, msg) from None
        if backlog is None:
            sock.listen()
        else:
            sock.listen(backlog)
        return sock
    except error:
        sock.close()
        raise



import os
import sys
from greenletio.io import wait_to_read, wait_to_write
from greenletio.patcher import copy_globals
from ssl import SSLWantReadError, SSLWantWriteError, PROTOCOL_TLS, Purpose, \
    CERT_NONE, CERT_REQUIRED, _ASN1Object
import ssl as _original_ssl_

copy_globals(_original_ssl_, globals())


class SSLContext(_original_ssl_.SSLContext):
    sslsocket_class = None

    def wrap_socket(self, sock, server_side=False,
                    do_handshake_on_connect=True,
                    suppress_ragged_eofs=True,
                    server_hostname=None, session=None):
        if hasattr(_original_ssl_.SSLSocket, '_create'):
            # python 3.7+
            return self.sslsocket_class._create(
                sock=sock,
                server_side=server_side,
                do_handshake_on_connect=do_handshake_on_connect,
                suppress_ragged_eofs=suppress_ragged_eofs,
                server_hostname=server_hostname,
                context=self,
                session=session
            )
        else:  # pragma: no cover
            # python 3.6
            return SSLSocket(sock=sock, server_side=server_side,
                             do_handshake_on_connect=do_handshake_on_connect,
                             suppress_ragged_eofs=suppress_ragged_eofs,
                             server_hostname=server_hostname,
                             _context=self, _session=session)


class SSLSocket(_original_ssl_.SSLSocket):
    if hasattr(_original_ssl_.SSLSocket, '_create'):
        @classmethod
        def _create(cls, *args, **kwargs):
            do_handshake_on_connect = kwargs.get(
                'do_handshake_on_connect', True)
            kwargs['do_handshake_on_connect'] = False
            sock = super(SSLSocket, cls)._create(*args, **kwargs)
            sock.do_handshake_on_connect = do_handshake_on_connect
            if sock._connected and do_handshake_on_connect:
                try:
                    sock.do_handshake()
                except (OSError, ValueError):  # pragma: no cover
                    sock.close()
                    raise
            return sock
    else:  # pragma: no cover
        def __init__(self, *args, **kwargs):
            do_handshake_on_connect = kwargs.get(
                'do_handshake_on_connect', True)
            kwargs['do_handshake_on_connect'] = False
            super().__init__(*args, **kwargs)
            self.do_handshake_on_connect = do_handshake_on_connect
            if self._connected and do_handshake_on_connect:
                self.do_handshake_on_connect = True
                try:
                    self.do_handshake()
                except (OSError, ValueError):
                    self.close()
                    raise

    def _nonblocking_io(self, method, *args, **kwargs):
        self.setblocking(False)
        while True:
            try:
                ret = method(*args, **kwargs)
                break
            except SSLWantReadError:
                wait_to_read(self.fileno())
            except SSLWantWriteError:  # pragma: no cover
                wait_to_write(self.fileno())
        return ret

    def do_handshake(self):
        return self._nonblocking_io(super().do_handshake)

    def accept(self, *args, **kwargs):
        wait_to_read(self.fileno())
        return self._nonblocking_io(super().accept, *args, **kwargs)

    def recv(self, *args, **kwargs):
        return self._nonblocking_io(super().recv, *args, **kwargs)

    def recv_into(self, *args, **kwargs):  # pragma: no cover
        return self._nonblocking_io(super().recv_into, *args, **kwargs)

    def send(self, *args, **kwargs):
        return self._nonblocking_io(super().send, *args, **kwargs)

    def sendall(self, data, flags=0):
        tail = self.send(data, flags)
        len_data = len(data)
        while tail < len_data:  # pragma: no cover
            tail += self.send(data[tail:], flags)

    def sendfile(self, *args, **kwargs):  # pragma: no cover
        raise RuntimeError('socket.sendfile is not supported')


SSLContext.sslsocket_class = SSLSocket


def create_default_context(purpose=Purpose.SERVER_AUTH, *, cafile=None,
                           capath=None, cadata=None):  # pragma: no cover
    """Create a SSLContext object with default settings.
    NOTE: The protocol and settings may change anytime without prior
          deprecation. The values represent a fair balance between maximum
          compatibility and security.
    """
    if not isinstance(purpose, _ASN1Object):
        raise TypeError(purpose)

    # SSLContext sets OP_NO_SSLv2, OP_NO_SSLv3, OP_NO_COMPRESSION,
    # OP_CIPHER_SERVER_PREFERENCE, OP_SINGLE_DH_USE and OP_SINGLE_ECDH_USE
    # by default.
    context = SSLContext(PROTOCOL_TLS)

    if purpose == Purpose.SERVER_AUTH:
        # verify certs and host name in client mode
        context.verify_mode = CERT_REQUIRED
        context.check_hostname = True

    if cafile or capath or cadata:
        context.load_verify_locations(cafile, capath, cadata)
    elif context.verify_mode != CERT_NONE:
        # no explicit cafile, capath or cadata but the verify mode is
        # CERT_OPTIONAL or CERT_REQUIRED. Let's try to load default system
        # root CA certificates for the given purpose. This may fail silently.
        context.load_default_certs(purpose)
    # OpenSSL 1.1.1 keylog file
    if hasattr(context, 'keylog_filename'):
        keylogfile = os.environ.get('SSLKEYLOGFILE')
        if keylogfile and not sys.flags.ignore_environment:
            context.keylog_filename = keylogfile
    return context


def wrap_socket(sock, keyfile=None, certfile=None,
                server_side=False, cert_reqs=CERT_NONE,
                ssl_version=PROTOCOL_TLS, ca_certs=None,
                do_handshake_on_connect=True,
                suppress_ragged_eofs=True,
                ciphers=None):  # pragma: no cover

    if server_side and not certfile:
        raise ValueError("certfile must be specified for server-side "
                         "operations")
    if keyfile and not certfile:
        raise ValueError("certfile must be specified")
    context = SSLContext(ssl_version)
    context.verify_mode = cert_reqs
    if ca_certs:
        context.load_verify_locations(ca_certs)
    if certfile:
        context.load_cert_chain(certfile, keyfile)
    if ciphers:
        context.set_ciphers(ciphers)
    return context.wrap_socket(
        sock=sock, server_side=server_side,
        do_handshake_on_connect=do_handshake_on_connect,
        suppress_ragged_eofs=suppress_ragged_eofs
    )



import asyncio
import collections
import weakref
import greenlet
from greenletio.core import await_, spawn
from greenletio.patcher import copy_globals
import threading as _original_threading_

copy_globals(_original_threading_, globals())

_active = {}


def get_ident():
    return id(greenlet.getcurrent())


class Lock:
    def __init__(self):
        self._lock = asyncio.Lock()
        self.locked = self._lock.locked
        self.release = self._lock.release

    def acquire(self, blocking=True, timeout=None):
        if not blocking and self.locked():
            return False
        if timeout is None or timeout < 0:
            return await_(self._lock.acquire())
        else:
            try:
                return await_(asyncio.wait_for(self._lock.acquire(), timeout))
            except asyncio.TimeoutError:
                return False

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args):
        return self.release()


class RLock:
    def __init__(self):
        self._lock = Lock()
        self.owner = None
        self.count = 0
        self.locked = self._lock.locked

    def __repr__(self):  # pragma: no cover
        owner = self.owner
        try:
            owner = _active[owner].name
        except KeyError:
            pass
        r = super().__repr__().split(' ', 1)
        return "<%s.%s %s, owner=%s, count=%d]>" % (
            self.__class__.__module__,
            self.__class__.__qualname__,
            r[1][:-2],
            owner,
            self.count)

    def acquire(self, blocking=True, timeout=-1):
        me = get_ident()
        if self.owner == me:
            self.count += 1
            return True
        ret = self._lock.acquire(blocking, timeout)
        if ret:
            self.owner = me
            self.count = 1
        return ret

    def release(self):
        if self.owner != get_ident():
            raise RuntimeError("cannot release un-acquired lock")
        self.count = self.count - 1
        if not self.count:
            self.owner = None
            self._lock.release()

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args):
        return self.release()


class Condition:
    def __init__(self, lock=None):
        if lock is None:
            lock = RLock()
        self._lock = lock
        self.locked = lock.locked
        self.acquire = lock.acquire
        self.release = lock.release
        self._waiters = collections.deque()

    def __repr__(self):  # pragma: no cover
        res = super().__repr__()
        extra = 'locked' if self.locked() else 'unlocked'
        if self._waiters:
            extra = f'{extra}, waiters:{len(self._waiters)}'
        return f'<{res[1:-1]} [{extra}]>'

    def wait(self, timeout=None):
        if not self.locked():
            raise RuntimeError('cannot wait on un-acquired lock')

        self.release()
        try:
            fut = asyncio.get_event_loop().create_future()
            self._waiters.append(fut)
            try:
                if timeout is None:
                    await_(fut)
                else:
                    try:
                        await_(asyncio.wait_for(fut, timeout))
                    except asyncio.TimeoutError:
                        return False
                return True
            finally:
                self._waiters.remove(fut)
        finally:
            cancelled = False
            while True:
                try:
                    self.acquire()
                    break
                except asyncio.CancelledError:  # pragma: no cover
                    cancelled = True
            if cancelled:  # pragma: no cover
                raise asyncio.CancelledError

    def wait_for(self, predicate, timeout=None):
        result = predicate()
        while not result:
            self.wait(timeout)
            result = predicate()
        return result

    def notify(self, n=1):
        if not self.locked():
            raise RuntimeError('cannot notify on un-acquired lock')

        idx = 0
        for fut in self._waiters:
            if idx >= n:  # pragma: no cover
                break

            if not fut.done():  # pragma: no cover
                idx += 1
                fut.set_result(False)

    def notify_all(self):
        self.notify(len(self._waiters))

    notifyAll = notify_all

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args):
        return self.release()


class Semaphore:
    def __init__(self, value=1):
        self._sem = asyncio.Semaphore(value)
        self.release = self._sem.release

    def acquire(self, blocking=True, timeout=None):
        if timeout is None:
            return await_(self._sem.acquire())
        else:
            try:
                return await_(asyncio.wait_for(self._sem.acquire(), timeout))
            except asyncio.TimeoutError:
                return False

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args):
        return self.release()


class BoundedSemaphore(Semaphore):
    def __init__(self, value=1):
        self._sem = asyncio.BoundedSemaphore(value)
        self.release = self._sem.release


class Event(asyncio.Event):
    def wait(self, timeout=None):
        if timeout is None:
            return await_(super().wait())
        else:
            try:
                return await_(asyncio.wait_for(super().wait(), timeout))
            except asyncio.TimeoutError:
                return False

    isSet = asyncio.Event.is_set


class Barrier(_original_threading_.Barrier):
    def __init__(self, parties, action=None, timeout=None):
        self._cond = Condition(Lock())
        self._action = action
        self._timeout = timeout
        self._parties = parties
        self._state = 0
        self._count = 0


class Thread(_original_threading_.Thread):
    def __init__(self, group=None, target=None, name=None, args=(),
                 kwargs=None, *, daemon=None):
        assert group is None, "group argument must be None for now"
        if kwargs is None:
            kwargs = {}
        self._target = target
        self._name = str(name or _original_threading_._newname())
        self._args = args
        self._kwargs = kwargs
        self._daemonic = False
        self._ident = None
        self._started = Event()
        self._ended = Event()
        self._is_stopped = False
        self._initialized = True
        self._task = None

    def start(self):
        if not self._initialized:  # pragma: no cover
            raise RuntimeError("thread.__init__() not called")

        if self._started.is_set():
            raise RuntimeError("threads can only be started once")

        self._task = spawn(self._bootstrap)
        self._started.set()

    def _set_ident(self):
        self._ident = get_ident()

    def _bootstrap(self):
        self._set_ident()
        _active[self._ident] = self
        try:
            self.run()
        finally:
            self._is_stopped = True
            del _active[self._ident]
            self._ended.set()

    def join(self, timeout=None):
        if not self._initialized:  # pragma: no cover
            raise RuntimeError("Thread.__init__() not called")
        if not self._started.is_set():
            raise RuntimeError("cannot join thread before it is started")
        if self is current_thread():
            raise RuntimeError("cannot join current thread")

        self._ended.wait()

    def is_alive(self):
        assert self._initialized, "Thread.__init__() not called"
        return not self._is_stopped and self._started.is_set()


class _DummyThread(Thread):

    def __init__(self):
        Thread.__init__(self, name=_original_threading_._newname("Dummy-%d"),
                        daemon=True)
        self._set_ident()
        _active[self._ident] = self

    def _stop(self):  # pragma: no cover
        pass

    def is_alive(self):  # pragma: no cover
        assert not self._is_stopped and self._started.is_set()
        return True

    def join(self, timeout=None):  # pragma: no cover
        assert False, "cannot join a dummy thread"


def current_thread():
    try:
        return _active[get_ident()]
    except KeyError:
        return _DummyThread()


class Timer(Thread):
    def __init__(self, interval, function, args=None, kwargs=None):
        Thread.__init__(self)
        self.interval = interval
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.finished = Event()

    def cancel(self):
        """Stop the timer if it hasn't finished yet."""
        self.finished.set()

    def run(self):
        self.finished.wait(self.interval)
        if not self.finished.is_set():
            self.function(*self.args, **self.kwargs)
        self.finished.set()


class _localbase(object):
    __slots__ = '_local__args', '_local__greens'

    def __new__(cls, *args, **kw):
        self = object.__new__(cls)
        object.__setattr__(self, '_local__args', (args, kw))
        object.__setattr__(self, '_local__greens', weakref.WeakKeyDictionary())
        if (args or kw) and \
                (cls.__init__ is object.__init__):  # pragma: no cover
            raise TypeError("Initialization arguments are not supported")
        return self


def _patch(thrl):
    greens = object.__getattribute__(thrl, '_local__greens')
    cur = greenlet.getcurrent()
    if cur not in greens:
        greens[cur] = {}
        cls = type(thrl)
        if cls.__init__ is not object.__init__:
            args, kw = object.__getattribute__(thrl, '_local__args')
            thrl.__init__(*args, **kw)
    object.__setattr__(thrl, '__dict__', greens[cur])


class local(_localbase):
    def __getattribute__(self, attr):
        _patch(self)
        return object.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        _patch(self)
        return object.__setattr__(self, attr, value)

    def __delattr__(self, attr):
        _patch(self)
        return object.__delattr__(self, attr)



from contextlib import contextmanager
import sys

patched = {}


def copy_globals(source_module, globals_dict):
    for attr in dir(source_module):
        if not getattr(globals_dict, attr, None):
            globals_dict[attr] = getattr(source_module, attr)


@contextmanager
def patch_blocking(modules=None):
    saved = {}
    saved_module_list = list(sys.modules.keys()).copy()
    if modules is None:
        modules = ['socket', 'ssl', 'threading', 'time']
    for module in modules:
        if module not in patched:
            patched[module] = getattr(
                __import__('greenletio.green.' + module).green, module)
    if '__greenletio_patched__' not in sys.modules:
        for module in modules:
            if module in sys.modules:
                saved[module] = sys.modules[module]
            sys.modules[module] = patched[module]
        sys.modules['__greenletio_patched__'] = True
    yield
    for module in saved:
        sys.modules[module] = saved[module]
    for module in list(sys.modules.keys()).copy():
        if module not in saved_module_list:
            del sys.modules[module]


def patch_psycopg2():
    import psycopg2
    from psycopg2.extensions import POLL_OK, POLL_READ, POLL_WRITE
    from greenletio.io import wait_to_read, wait_to_write

    if not hasattr(psycopg2.extensions,
                   'set_wait_callback'):  # pragma: no cover
        raise ImportError(
            "support for coroutines not available in this Psycopg version (%s)"
            % psycopg2.__version__)

    def psycopg2_wait_callback(conn):  # pragma: no cover
        fd = conn.fileno()
        while True:
            state = conn.poll()
            if state == POLL_OK:
                break
            elif state == POLL_READ:
                wait_to_read(fd)
            elif state == POLL_WRITE:
                wait_to_write(fd)
            else:
                raise psycopg2.OperationalError(
                    "Bad result from poll: %r" % state)

    psycopg2.extensions.set_wait_callback(psycopg2_wait_callback)



import unittest
import pytest
from greenletio.core import bridge
from greenletio.green import threading, time


class TestThreading(unittest.TestCase):
    def setUp(self):
        bridge.reset()

    def tearDown(self):
        bridge.stop()

    def test_lock(self):
        var = None
        lock = threading.Lock()

        def t1():
            nonlocal var
            with pytest.raises(RuntimeError):
                lock.release()
            with lock:
                time.sleep(0.05)
                var = 'foo'

        def t2():
            nonlocal var
            time.sleep(0.02)
            assert lock.acquire(blocking=False) is False
            assert lock.acquire(timeout=0.01) is False
            with lock:
                assert var == 'foo'
                var = 'bar'

        th1 = threading.Thread(target=t1)
        th2 = threading.Thread(target=t2)
        th1.start()
        th2.start()
        th1.join()
        th2.join()
        assert var == 'bar'

    def test_rlock(self):
        var = None
        lock = threading.RLock()

        def t1():
            nonlocal var
            with lock:
                with lock:
                    var = 'foo'
                    time.sleep(0.05)

        def t2():
            nonlocal var
            time.sleep(0.02)
            assert lock.acquire(blocking=False) is False
            with pytest.raises(RuntimeError):
                lock.release()
            with lock:
                with lock:
                    assert var == 'foo'
                    var = 'bar'

        th1 = threading.Thread(target=t1)
        th2 = threading.Thread(target=t2)
        th1.start()
        th2.start()
        th1.join()
        th2.join()
        assert var == 'bar'

    def test_condition(self):
        var = None
        cv = threading.Condition()

        def consumer():
            nonlocal var
            with pytest.raises(RuntimeError):
                cv.wait()
            with cv:
                assert cv.wait(timeout=0) is False
                cv.wait()
                assert var == 'foo'
                var = 'bar'

        def producer():
            nonlocal var
            with pytest.raises(RuntimeError):
                cv.notify()
            time.sleep(0.05)
            with cv:
                var = 'foo'
                cv.notify_all()

        th1 = threading.Thread(target=consumer)
        th2 = threading.Thread(target=producer)
        th1.start()
        th2.start()
        th1.join()
        th2.join()
        assert var == 'bar'

    def test_semaphore(self):
        var = None
        sem = threading.Semaphore()

        def t():
            nonlocal var
            for i in range(10):
                with sem:
                    assert sem.acquire(timeout=0.01) is False
                    assert var is None
                    var = 'foo'
                    time.sleep(0)
                    var = None

        ths = [threading.Thread(target=t) for i in range(5)]
        for th in ths:
            th.start()
        for th in ths:
            th.join()
        assert var is None

    def test_bounded_semaphore(self):
        var = 0
        sem = threading.BoundedSemaphore(2)

        def t():
            nonlocal var
            for i in range(10):
                with sem:
                    assert var <= 1
                    var += 1
                    time.sleep(0)
                    var -= 1

        ths = [threading.Thread(target=t) for i in range(5)]
        for th in ths:
            th.start()
        for th in ths:
            th.join()
        assert var == 0
        with pytest.raises(ValueError):
            sem.release()

    def test_event(self):
        var = None
        ev = threading.Event()

        def t():
            nonlocal var
            time.sleep(0.05)
            var = 'foo'
            ev.set()

        th = threading.Thread(target=t)
        th.start()
        assert not ev.is_set()
        ev.wait()
        assert ev.is_set()
        assert var == 'foo'

    def test_timer(self):
        var = None

        def t(arg):
            nonlocal var
            var = arg

        tm = threading.Timer(0.05, t, (42,))
        tm.start()
        assert var is None
        tm.join()
        assert var == 42

    def test_timer_canceled(self):
        var = None

        def t(arg):
            nonlocal var
            var = arg

        tm = threading.Timer(0.05, t, (42,))
        tm.start()
        tm.cancel()
        assert var is None
        tm.join()
        assert var is None

    def test_barrier(self):
        var = [0, 0, 0, 0, 0, 0]
        b = threading.Barrier(6)

        def t():
            time.sleep(0.05)
            ret = b.wait()
            var[ret] = 1

        ths = [threading.Thread(target=t) for i in range(5)]
        for th in ths:
            th.start()
        ret = b.wait()
        for th in ths:
            th.join()
        for i in range(len(var)):
            assert var[i] == 0 if i == ret else 1

    def test_local(self):
        local = threading.local()

        def t(n):
            local.var = n
            time.sleep(0.05)
            assert local.var == n

        local.var = 'foo'
        ths = [threading.Thread(target=t, args=(i,)) for i in range(5)]
        for th in ths:
            th.start()
        for th in ths:
            th.join()
        assert local.var == 'foo'
        del local.var
        with pytest.raises(AttributeError):
            local.var

    def test_local_object(self):
        class MyLocal(threading.local):
            def __init__(self, n):
                self.var = n

        local = MyLocal('foo')

        def t(n):
            assert local.var == 'foo'
            local.var = n
            time.sleep(0.05)
            assert local.var == n

        local.var = 'bar'
        ths = [threading.Thread(target=t, args=(i,)) for i in range(5)]
        for th in ths:
            th.start()
        for th in ths:
            th.join()
        assert local.var == 'bar'
        del local.var
        with pytest.raises(AttributeError):
            local.var

    def test_thread(self):
        def t(foo, bar=None):
            with pytest.raises(RuntimeError):
                threading.current_thread().join()
            assert foo == 'foo'
            assert bar == 'bar'

        th = threading.Thread(target=t, args=('foo',), kwargs={'bar': 'bar'})
        with pytest.raises(RuntimeError):
            th.join()
        th.start()
        assert th.is_alive()
        th.join()
        assert th.is_alive() is False
        with pytest.raises(RuntimeError):
            th.start()



import asyncio
import unittest
import pytest
from greenletio import async_, await_, spawn
from greenletio.core import bridge


class TestCore(unittest.TestCase):
    def setUp(self):
        bridge.reset()

    def tearDown(self):
        bridge.stop()

    def test_await_function_with_internal_loop(self):
        var = None

        async def a(arg):
            nonlocal var
            var = arg
            return arg

        ret = await_(a(42))
        assert ret == 42
        assert var == 42

        bridge.stop()
        assert not bridge.running
        assert not asyncio.get_event_loop().is_running()

    def test_await_decorator_with_internal_loop(self):
        var = None

        @await_
        async def a(arg):
            nonlocal var
            var = arg
            return arg

        ret = a(42)
        assert ret == 42
        assert var == 42

        bridge.stop()
        assert not bridge.running
        assert not asyncio.get_event_loop().is_running()

    def test_async_await_with_external_loop(self):
        var = None

        async def a(arg):
            nonlocal var
            var = arg
            return arg

        @async_
        def b(arg):
            return await_(a(arg))

        @async_
        def c(arg):
            return arg

        async def d(arg):
            assert await(c(arg)) == arg
            return await b(arg)

        ret = asyncio.get_event_loop().run_until_complete(d(42))
        assert ret == 42
        assert var == 42

    def test_async_await_exception(self):
        @async_
        def a(arg):
            raise RuntimeError(arg)

        async def b(arg):
            with pytest.raises(RuntimeError) as error:
                await a(arg)
            assert type(error.value) == RuntimeError and \
                str(error.value) == '42'

        asyncio.get_event_loop().run_until_complete(b(42))

    def test_async_await_exception2(self):
        async def a(arg):
            raise RuntimeError(arg)

        @async_
        def b(arg):
            with pytest.raises(RuntimeError) as error:
                await_(a(arg))
            assert type(error.value) == RuntimeError and \
                str(error.value) == '42'

        async def c(arg):
            await b(arg)

        asyncio.get_event_loop().run_until_complete(c(42))

    def test_gather_with_external_loop(self):
        var = 0

        @async_
        def a():
            nonlocal var
            await_(asyncio.sleep(0.01))
            var += 1
            return var

        @async_
        def b():
            nonlocal var
            await_(asyncio.sleep(0.01))
            var += 2
            return var

        async def c():
            tasks = [a(), b()]
            return await asyncio.gather(*tasks)

        ret = asyncio.get_event_loop().run_until_complete(c())
        assert ret == [1, 3] or ret == [3, 2]

    def test_await_raises_exception(self):
        async def a(arg):
            raise RuntimeError('foo')

        with pytest.raises(RuntimeError) as exc:
            await_(a(42))
        assert exc.type == RuntimeError
        assert str(exc.value) == 'foo'

    def test_async_raises_exception(self):
        @async_
        def a(arg):
            raise RuntimeError('foo')

        async def b():
            with pytest.raises(RuntimeError) as exc:
                await a(42)
            assert exc.type == RuntimeError
            assert str(exc.value) == 'foo'

        asyncio.get_event_loop().run_until_complete(b())

    def test_spawn(self):
        var = 0

        def a(arg):
            nonlocal var
            var += arg

        def b(arg):
            nonlocal var
            var += arg

        async def c():
            ga = spawn(a, 40)
            gb = spawn(b, 2)
            while not ga.dead or not gb.dead:
                await asyncio.sleep(0)

        asyncio.get_event_loop().run_until_complete(c())
        assert var == 42



from greenlet import getcurrent
from greenletio.core import bridge


def _reader_callback(fd, gl):
    bridge.loop.remove_reader(fd)
    bridge.schedule(gl)


def _writer_callback(fd, gl):
    bridge.loop.remove_writer(fd)
    bridge.schedule(gl)


def wait_to_read(fd):
    if not bridge.running and not bridge.starting:  # pragma: no cover
        bridge.start()
    bridge.loop.add_reader(fd, _reader_callback, fd, getcurrent())
    bridge.switch()


def wait_to_write(fd):
    if not bridge.running and not bridge.starting:  # pragma: no cover
        bridge.start()
    bridge.loop.add_writer(fd, _writer_callback, fd, getcurrent())
    bridge.switch()



import asyncio
import atexit
from collections import deque
import functools
import sys
from greenlet import greenlet, getcurrent


class GreenletBridge:
    def __init__(self):
        self.reset()
        self.wait_event = None

    def reset(self):
        self.starting = False
        self.running = False
        self.stopping = False
        self.loop = None
        self.bridge_greenlet = None
        self.scheduled = deque()
        self.pending_io = object()

    def schedule(self, gl, *args, **kwargs):
        self.scheduled.append((gl, args, kwargs))
        if self.wait_event:
            self.wait_event.set()

    def run(self):
        async def async_run():
            if self.bridge_greenlet != getcurrent():
                self.bridge_greenlet = getcurrent()
            self.starting = False
            self.stopping = False
            self.running = True
            while not self.stopping or self.scheduled:
                self.wait_event.clear()
                while self.scheduled:
                    gl, args, kwargs = self.scheduled.popleft()
                    self.loop.create_task(_run_greenlet_in_aio(
                        gl, args=args, kwargs=kwargs))
                if self.stopping:  # pragma: no cover
                    break
                await self.wait_event.wait()
            self.running = False

        # get the asyncio loop
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:  # pragma: no cover
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        self.loop = loop
        self.wait_event = asyncio.Event()
        if not loop.is_running():
            # neither the loop nor the bridge are running
            # start a loop with the bridge task
            loop.run_until_complete(async_run())
        else:
            # the loop is already running, but the bridge isn't
            # start the bridge as a task
            loop.create_task(async_run())

    def start(self):
        assert not self.running and not self.starting
        atexit.register(self.stop)
        self.reset()
        self.starting = True
        self.schedule(getcurrent())
        self.switch()

    def stop(self):
        if self.running:
            self.stopping = True
            self.wait_event.set()
            if self.bridge_greenlet != getcurrent():
                self.bridge_greenlet.parent = getcurrent()
                while not self.bridge_greenlet.dead:  # pragma: no cover
                    self.bridge_greenlet.switch()

    def switch(self):
        if self.bridge_greenlet is None:
            self.bridge_greenlet = greenlet(self.run)
        if self.wait_event:
            self.wait_event.set()
        return self.bridge_greenlet.switch()


bridge = GreenletBridge()


async def _run_greenlet_in_aio(gl, args=None, kwargs=None):
    coro = gl.switch(*args, **kwargs)
    while gl and coro != ():
        ret = None
        try:
            ret = await coro
        except:  # noqa: E722
            coro = gl.throw(*sys.exc_info())
        else:
            coro = gl.switch(ret)
    return coro


def async_(fn):
    @functools.wraps(fn)
    def decorator(*args, **kwargs):
        if not bridge.running and not bridge.starting:
            bridge.start()

        async def coro(fn, *args, **kwargs):
            future = asyncio.Future()

            def gl(future, fn, *args, **kwargs):
                try:
                    future.set_result(fn(*args, **kwargs))
                except:  # noqa: E722
                    future.set_exception(sys.exc_info()[1])

            bridge.schedule(greenlet(gl), future, fn, *args, **kwargs)
            return await future

        return coro(fn, *args, **kwargs)

    return decorator


def await_(coro_or_fn):
    if asyncio.iscoroutine(coro_or_fn) or asyncio.isfuture(coro_or_fn):
        # we were given an awaitable --> await it
        if not bridge.running and not bridge.starting:
            bridge.start()

        return bridge.bridge_greenlet.switch(coro_or_fn)
    else:
        # assume decorator usage
        @functools.wraps(coro_or_fn)
        def decorator(*args, **kwargs):
            return await_(coro_or_fn(*args, **kwargs))

        return decorator


def spawn(fn, *args, **kwargs):
    if not bridge.running and not bridge.starting:
        bridge.start()

    def _fn(*args, **kwargs):
        getcurrent().parent = bridge.bridge_greenlet
        fn(*args, **kwargs)

    gl = greenlet(_fn)
    bridge.schedule(gl, *args, **kwargs)
    return gl



import asyncio
from greenletio.core import await_
from greenletio.patcher import copy_globals
import time as _original_time_

copy_globals(_original_time_, globals())


def sleep(seconds):  # pragma: no cover
    await_(asyncio.sleep(seconds))



import asyncio
import unittest
from greenletio import spawn
from greenletio.core import bridge
from greenletio.green import socket, ssl

# Tests in this module use server and client certificates
#
# To generate server certificate:
# openssl req -x509 -newkey rsa:4096 -keyout key.pem -out cert.pem -days 365
# -nodes -subj "/CN=example.com"
#
# To generate client certificate:
# openssl req -x509 -newkey rsa:4096 -keyout client.key -out client.crt
# -days 365 -nodes -subj "/CN=example.com"


class TestSSL(unittest.TestCase):
    def setUp(self):
        bridge.reset()

    def tearDown(self):
        bridge.stop()

    def test_sendall_recv(self):
        var = None

        def server():
            server_socket = socket.socket()
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(('127.0.0.1', 7000))
            server_socket.listen(5)
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            context.load_cert_chain('tests/server.crt', 'tests/server.key')
            context.load_verify_locations('tests/client.crt')
            ssl_socket = context.wrap_socket(server_socket, server_side=True)
            conn, _ = ssl_socket.accept()
            data = conn.recv(1024)
            conn.sendall(data.upper())
            conn.close()
            ssl_socket.close()

        def client():
            nonlocal var
            client_socket = socket.socket()
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH,
                                                 cafile='tests/server.crt')
            context.load_cert_chain('tests/client.crt', 'tests/client.key')
            ssl_socket = context.wrap_socket(client_socket,
                                             server_hostname='example.com')
            ssl_socket.connect(('127.0.0.1', 7000))
            ssl_socket.sendall(b'hello')
            var = ssl_socket.recv(1024)
            ssl_socket.close()

        async def main():
            nonlocal var
            spawn(server)
            spawn(client)
            while var is None:
                await asyncio.sleep(0)

        asyncio.get_event_loop().run_until_complete(main())
        assert var == b'HELLO'



import platform
import socket
import sys
import unittest
from greenletio import patch_blocking, patch_psycopg2
from greenletio.green import socket as green_socket, \
    threading as green_threading


class TestPatcher(unittest.TestCase):
    def test_patch(self):
        assert 'socketserver' not in sys.modules
        if 'threading' in sys.modules:
            del sys.modules['threading']
        with patch_blocking():
            assert sys.modules['socket'] == green_socket
            assert sys.modules['threading'] == green_threading
            import socketserver
            assert socketserver.socket == green_socket
            assert '__greenletio_patched__' in sys.modules
        assert sys.modules['socket'] == socket
        assert 'threading' not in sys.modules
        assert 'socketserver' not in sys.modules
        assert '__greenletio_patched__' not in sys.modules

    def test_patch_partial(self):
        assert 'socketserver' not in sys.modules
        original_threading = sys.modules.get('threading')
        with patch_blocking(['socket']):
            assert sys.modules['socket'] == green_socket
            assert sys.modules.get('threading') == original_threading
            import socketserver
            assert socketserver.socket == green_socket
            assert '__greenletio_patched__' in sys.modules
        assert sys.modules['socket'] == socket
        assert sys.modules.get('threading') == original_threading
        assert 'socketserver' not in sys.modules
        assert '__greenletio_patched__' not in sys.modules

    def test_patch_recursive(self):
        assert 'socketserver' not in sys.modules
        if 'threading' in sys.modules:
            del sys.modules['threading']
        with patch_blocking():
            with patch_blocking():
                assert sys.modules['socket'] == green_socket
                assert sys.modules['threading'] == green_threading
                import socketserver
                assert socketserver.socket == green_socket
                assert '__greenletio_patched__' in sys.modules
            assert sys.modules['socket'] == green_socket
            assert sys.modules['threading'] == green_threading
            assert 'socketserver' not in sys.modules
            assert '__greenletio_patched__' in sys.modules
        assert sys.modules['socket'] == socket
        assert 'threading' not in sys.modules
        assert 'socketserver' not in sys.modules
        assert '__greenletio_patched__' not in sys.modules

    @unittest.skipIf(platform.python_implementation() != 'CPython',
                     'psycopg2-binary does not install on pypy3')
    def test_patch_psycopg2(self):
        import psycopg2
        assert psycopg2.extensions.get_wait_callback() is None
        patch_psycopg2()
        assert psycopg2.extensions.get_wait_callback() is not None
